import argparse
import sys
from exifScan import dualLogging, watchDirectory
from exifScan.runExifScan import ExifScanner
import json
import logging

import os


def getArgs():
    parser = argparse.ArgumentParser(
        prog='Watch Directory - PhotoDatabase',
        description='This script runs watchdog on a folder after taking a directory snapshot, and then keeps a geopackage map in sync with changes while the script is active.',
        epilog='I hope this proves useful.')

    parser.add_argument('path', help='The path to scan for Exif data. This argument is required')
    parser.add_argument('--WorkDirectory', default='.', help='The working directory. will default to cwd.')
    parser.add_argument('--MetashapeDirectory', help='The Metashape directory. Will search this directory recursively to find Metashape project files. Will only work with python <3.11. Will default to ScanFolder.')
    parser.add_argument('--KartRemote', help='The Kart remote page. Will push to the remote repo if set')
    parser.add_argument('--KartDirectory', help='The Kart directory. Will write to a kart repo if set')
    parser.add_argument('--GeopackageDirectory', help='The output directory. Will default to WorkDirectory.')
    parser.add_argument('--LogDirectory', default='.', help='The log directory.')
    parser.add_argument('--OutputGeopackage', default='Photos.gpkg', help='The output GeoPackage')
    parser.add_argument('--htmlFolder', help='The HTML folder. Will output a html map and table to this folder if set.')
    parser.add_argument('--mapFileName', default='map.html', help='The map file name. Only works if --htmlFolder is set')
    parser.add_argument('--otherMapLayers', nargs='+', type=str, help='A list of paths representing other layers that will be loaded by geopandas to display in the html Map.')
    parser.add_argument('--otherMapLayerNames', nargs='+', type=str, help='A list of layer names representing other layers that will be loaded by geopandas to display in the html Map.')
    parser.add_argument('--ExifOutputFile', default="outdb", help='The Exif output file name in the WorkDirectory Do not add file extension, will automatically append json extension.')
    parser.add_argument('--readExifFromOutputFile', action='store_true', default=False, help='Whether to read Exif from output file. Uses the --ExifOutputFile to find the file.')
    parser.add_argument('--recursive', action='store_true', default=False, help='Whether to run recursively. Will not attempt to remove files from existing geodf')
    parser.add_argument('--ignoreDirs', nargs='+', default=[], help='Directories that watchdog will not scan after the initial scan.')
    parser.add_argument('--ignoreMetashape', action='store_true', help='ignore file directories that include the string ".files" and "auto-project".')
    parser.add_argument('--persistObserver', action='store_true', default=True, help='Persist the Observer once stopped by saving the a pickle file of the last image.')
    parser.add_argument('--dualLogging', action='store_true', default=False, help='Whether logging should log to console and to files.')
    parser.add_argument("--groupByTags", nargs='+', help="List of tags to group by", default=["Make", "Model"])
    parser.add_argument("--xyTagNames", nargs='+', help="List of tag names for x and y coordinates", default=['GPSLongitude', 'GPSLatitude'])
    parser.add_argument("--zTagName", type=str, help="Tag name for z coordinate", default='GPSAltitude')
    parser.add_argument("--imageFileExtensions", nargs='+', help="Image file extensions", default=['JPG'])
    parser.add_argument("--metashapeCacheTime", help="Time to keep Metashape data cached (in minutes)", default=120)
    parser.add_argument("--pollingInterval", help="Time (in seconds) between polling intervals. Will revert to default if unset.")
    parser.add_argument("--config", type=str, help="The config file, a path to a json file.")
    parser.add_argument("--watchDirectory", action='store_true', default=False, help='Whether to watch direcotry for changes after initial scan.')
    parser.add_argument('--logLevel', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'], default='INFO', help='Set the logging level')
    parser.add_argument("--timerInterval", help="Time between change events before processing changes", default=30)
    parser.add_argument("--linzAPIKey", help="Time between change events before processing changes", default=None)
    args = parser.parse_args()

    if args.config:
        try:
            with open(args.config, 'r') as f:
                config_args = json.load(f)
            for key, value in vars(args).items():
                defaultValue = parser.get_default(key)
                if value != defaultValue:
                    config_args[key] = value
                else:
                    if key not in config_args:
                        config_args[key] = value

            args = argparse.Namespace(**config_args)
        except Exception as e:
            print(f'An error occurred with the config file: {e}')
            sys.exit()
        delattr(args, 'config')

    return args


def main():

    args = getArgs()
    argsDict = vars(args)
    path = argsDict['path']
    prettyNameSpace = '\n'.join(f'{k}: {v}' for k, v in argsDict.items())
    path = argsDict.pop('path', None)
    enable_dualLogging = argsDict.pop('dualLogging', None)
    logLevel = argsDict.pop('logLevel', None)
    log_directory = argsDict.pop('LogDirectory', '.')

    if enable_dualLogging:

        dualLogging.windows_enable_ansi_terminal_mode()
        logfile = os.path.join(log_directory, 'runExifScan.log')
        dualLogging.set_up_logging(console_log_output="stdout", console_log_level="info", console_log_color=True,
                                   logfile_file=logfile, logfile_log_level="debug", logfile_log_color=False,
                                   warn_file=logfile + "warn.log", warn_log_level="warning", warn_log_color=False,
                                   log_line_template="%(color_on)s[%(asctime)s] [%(threadName)s] [%(levelname)-8s] %(message)s%(color_off)s")
        logging.info(f'Logfile path: {logfile}')

    else:
        logging.basicConfig(format='[%(asctime)s] [%(threadName)s] %(levelname)s:%(message)s', level=getattr(logging, logLevel))

    logging.warning('======================================================================')
    # pretty each arg on a newline
    logging.debug(f"Arguments used for watchDirectory:{prettyNameSpace}")
    watch = argsDict.pop('watchDirectory', None)
    if watch:
        watchDirectory.watchFolder(path, argsDict)
    else:
        scanner = ExifScanner(**argsDict)
        scanner.scan_for_photos(path)


if __name__ == "__main__":
    main()



#Original plugin for Leaflet by GreenInfo-Network
#Port to folium by Carlos Charletti
from branca.element import MacroElement
from jinja2 import Template

from folium.elements import JSCSSMixin
from folium.utilities import parse_options


class Creditos(JSCSSMixin, MacroElement):
 

    _template = Template(
        """
        {% macro script(this, kwargs) %}
            L.controlCredits(
                {{ this.options|tojson }}
            ).addTo({{this._parent.get_name()}});
        {% endmacro %}
        """
    )  # noqa

    default_js = [
        (
            "leaflet-control-credits.js",
            "https://cdn.jsdelivr.net/gh/carlosign/Folium.ControlCredits-Plugin@main/js/leaflet-control-credits.js",
        )
    ]
    default_css = [
        (
            "leaflet-control-credits.css",
            "https://cdn.jsdelivr.net/gh/carlosign/Folium.ControlCredits-Plugin@main/js/leaflet-control-credits.css",
        )
    ]

    def __init__(
        self,
        imageurl="",
        imagealt="",
        tooltip="",
        width="45px",
        height="45px",
        expandcontent= "",
        **kwargs
    ):
        super().__init__()
        self._name = "Creditos"
        self.options = parse_options(
            imageurl=imageurl,
            imagealt=imagealt,
            tooltip=tooltip,
            width=width,
            height=height,
            expandcontent=expandcontent,
            **kwargs
        )



from __future__ import annotations
from watchdog.observers.api import DEFAULT_EMITTER_TIMEOUT, DEFAULT_OBSERVER_TIMEOUT, BaseObserver, EventEmitter, ObservedWatch
from watchdog.events import (
    DirCreatedEvent,
    DirDeletedEvent,
    DirModifiedEvent,
    DirMovedEvent,
    FileCreatedEvent,
    FileDeletedEvent,
    FileModifiedEvent,
    FileMovedEvent,
)

from watchdog.utils.dirsnapshot import DirectorySnapshot, DirectorySnapshotDiff, EmptyDirectorySnapshot
import threading
import pickle
import os
import logging
from datetime import datetime as dt


class PersistentPollingEmitter(EventEmitter):
    """
    Platform-independent emitter that polls a directory to detect file
    system changes.
    """

    def __init__(
        self,
        event_queue,
        watch,
        event_handler,
        previous_snapshot_file,
        on_snapshot=None,
        timeout=DEFAULT_EMITTER_TIMEOUT,
        event_filter=None,
        stat=os.stat,
        listdir=os.scandir,
    ):
        super().__init__(event_queue, watch, timeout, event_filter)
        previous_snapshots = load_pickle_file(previous_snapshot_file)
        if previous_snapshots is None:
            previous_snapshots = dict()
        self._hasPreviousSnapshot = isinstance(previous_snapshots.get(self.watch.path, None), DirectorySnapshot)
        self._snapshot: DirectorySnapshot = previous_snapshots.get(self.watch.path, EmptyDirectorySnapshot())
        self._event_handler = event_handler
        self._lock = threading.Lock()
        self.on_snapshot = on_snapshot
        self._take_snapshot = lambda: DirectorySnapshot(
            self.watch.path, self.watch.is_recursive, stat=stat, listdir=listdir
        )

    def on_thread_start(self):

        new_snapshot = self._take_snapshot()

        if not self._hasPreviousSnapshot:

            logging.info('Invalid previous snapshot.')
            self._snapshot = new_snapshot
        else:
            logging.info('Using previous snapshot.')
            self._event_handler.recentSnapshot = self._snapshot
            self._event_handler.currentSnapshot = self._snapshot
            self._event_handler.update_snapshot(new_snapshot)
            # diff immediately?

    def get_handler_snapshot(self):
        return self._event_handler.get_snapshot()

    def queue_events(self, timeout):
        # We don't want to hit the disk continuously.
        # timeout behaves like an interval for polling emitters.
        if self.stopped_event.wait(timeout):
            return

        with self._lock:
            if not self.should_keep_running():
                return
            logging.debug("Starting Diff.")

            # Get event diff between fresh snapshot and previous snapshot.
            # Update snapshot.
            try:
                snapshotStart = dt.now()
                new_snapshot = self._take_snapshot()
                self._event_handler.update_snapshot(new_snapshot)
                if self.on_snapshot:
                    self.on_snapshot(dt.now() - snapshotStart)

            except OSError:
                self.queue_event(DirDeletedEvent(self.watch.path))
                self.stop()
                return

            events = DirectorySnapshotDiff(self._snapshot, new_snapshot)
            self.old_snapshot = self._snapshot
            self._snapshot = new_snapshot

            # Files.
            for src_path in events.files_deleted:
                self.queue_event(FileDeletedEvent(src_path))
            for src_path in events.files_modified:
                self.queue_event(FileModifiedEvent(src_path))
            for src_path in events.files_created:
                self.queue_event(FileCreatedEvent(src_path))
            for src_path, dest_path in events.files_moved:
                self.queue_event(FileMovedEvent(src_path, dest_path))

            # Directories.
            for src_path in events.dirs_deleted:
                self.queue_event(DirDeletedEvent(src_path))
            for src_path in events.dirs_modified:
                self.queue_event(DirModifiedEvent(src_path))
            for src_path in events.dirs_created:
                self.queue_event(DirCreatedEvent(src_path))
            for src_path, dest_path in events.dirs_moved:
                self.queue_event(DirMovedEvent(src_path, dest_path))


def load_pickle_file(file_path):
    if not os.path.exists(file_path):
        return None

    try:
        with open(file_path, 'rb') as f:
            data = pickle.load(f)
            return data
    except (pickle.UnpicklingError, EOFError, AttributeError, ImportError, IndexError) as e:
        print(f"Error loading pickle file: {e}")
        return None


class PersistentPollingObserver(BaseObserver):
    """
    Platform-independent observer that polls a directory to detect file
    system changes.
    """

    def __init__(self, save_to, on_snapshot=None, protocol=0, timeout=DEFAULT_OBSERVER_TIMEOUT):
        self._filename = save_to
        self.on_snapshot = on_snapshot
        self._protocol = protocol
        self.previous_snapshot_exists = os.path.exists(self._filename)
        super().__init__(emitter_class=PersistentPollingEmitter, timeout=timeout)

    def schedule(self, event_handler, path, recursive=False, event_filter=None):
        with self._lock:
            watch = ObservedWatch(path, recursive, event_filter)
            self._add_handler_for_watch(event_handler, watch)

            # If we don't have an emitter for this watch already, create it.
            if self._emitter_for_watch.get(watch) is None:
                emitter = self._emitter_class(event_queue=self.event_queue, watch=watch, event_handler=event_handler, previous_snapshot_file=self._filename, timeout=self.timeout,
                                              event_filter=event_filter, on_snapshot=self.on_snapshot)
                if self.is_alive():
                    emitter.start()
                self._add_emitter(emitter)
            self._watches.add(watch)
        return watch

    def stop(self, *args, **kwargs):
        snapshots = {emitter.watch.path: emitter.get_handler_snapshot() for emitter in self.emitters}
        dir_name = os.path.dirname(self._filename)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
            logging.warning(f'Created folder {dir_name} to dump DirectorySnapshot.')
        with open(self._filename, 'wb') as f:
            pickle.dump(snapshots, f, self._protocol)
        logging.info(f'Saved pickled DirectorySnapshot at {self._filename}.')
        super().stop(*args, **kwargs)



import os
import sys
import logging
not_supported = False
if (sys.version_info[1] > 11):
    logging.error(f'Metashape not supported in python 3.11 or greater. This system appears to be running Python {sys.version_info[0]}.{sys.version_info[1]}')
    not_supported = True

try:
    import Metashape
except ImportError:
    logging.error("Metashape is not installed. Please install the package to proceed.")
    not_supported = True


def main(Directory):

    if not_supported:
        return

    ScannedDir = Directory
    ProjectPathList = []
    logging.info("Searching for Metashape project(.psx and .psz) files.")
    for root, _, files in os.walk(ScannedDir):
        for file in files:
            if file.endswith((".psx", ".psz")):
                FilePath = (os.path.join(root, file))
                ProjectPathList.append(FilePath)
                logging.debug(f'Added Project :{FilePath}')
    projectNo = len(ProjectPathList)
    logging.info(f"Located {projectNo} psx and psz files.")

    if (projectNo == 0):
        return None

    Doc = Metashape.Document()
    CameraDirDict = {}
    for ProjectPath in ProjectPathList:
        try:
            # read-only to ensure that no data is mutated.
            Doc.open(ProjectPath, read_only=True)
            logging.debug(f'Read {ProjectPath}')
            Chunks = Doc.chunks
            for Chunk in Chunks:
                CameraList = Chunk.cameras
                for Camera in CameraList:
                    if Camera.photo is not None:
                        try:
                            # get photo directory, and add directory to list if it isn't there.
                            CameraPath = Camera.photo.path
                            
                            CameraDir = CameraPath.rpartition('/')[0]
                            
                            if CameraDir not in CameraDirDict:
                                CameraDirDict[CameraDir] = []
                            # replace backslash with forwardslash to make paths consistent.
                            sanitised = ProjectPath.replace("\\", "/")
                            if sanitised not in CameraDirDict[CameraDir]:
                                CameraDirDict[CameraDir].append(sanitised)
                                logging.debug(f'Added ProjectPath {sanitised} to {CameraDir.encode("utf-8")}')
                        except Exception as e:
                            logging.error(f"Unable to read {Camera} in project {ProjectPath} . Error: {e}")
                    else:
                        logging.debug(f'Project:{ProjectPath}: Camera {Camera}has no attribute "path"')
        except Exception as e:
            logging.error(f"Problem with project: {ProjectPath}. Error: {e}")
    logging.info(f"Discovered {len(CameraDirDict)} Metashape project files that contained images from the database.")

    # list are object dtypes, so converted to strings when saving in pyogrio. Converting here to control how it is joined.
    # Maybe use a supported dtype in the future?
    for key in CameraDirDict:
        CameraDirDict[key] = ",\n".join(CameraDirDict[key])

    return CameraDirDict



import pandas as pd
import pickle
import os
from datetime import datetime as dt, timedelta as td
import logging
from exifScan.metashape import PsxLocator


class MetadataWriter:
    def __init__(self, WorkDirectory, metashapeCacheTime, MetashapeDirectory):
        self.WorkDirectory = WorkDirectory
        self.metashapeCacheTime = metashapeCacheTime
        self.MetashapeDirectory = MetashapeDirectory
        self.metashape_cache_file = os.path.join(self.WorkDirectory, 'metashapeCache.pkl')

    def update_metadata(self, gpdDf):
        startTime = dt.now()

        gpdDf = self.locate_and_match_metadata(gpdDf)
        gpdDf = self.associate_metashape_projects(gpdDf)

        logging.info(f"The Metadata was created in {dt.now() - startTime} (h:mm:ss).")
        return gpdDf

    def locate_and_match_metadata(self, gpdDf):
        FileName = "metadata.xlsx"
        not_found = []
        fileNo = 0

        for folder in gpdDf["SourceFileDir"].unique():
            cur_dir = folder
            try:
                while True:
                    file_list = os.listdir(cur_dir)
                    parent_dir = os.path.dirname(cur_dir)
                    if FileName in file_list:
                        fileNo += 1
                        ExcelFile = os.path.join(cur_dir, FileName)
                        logging.debug(f"Metadata file {FileName} for {folder} found in {cur_dir}")
                        break
                    else:
                        ExcelFile = None
                        if cur_dir == parent_dir:
                            logging.warning(f"Metadata file {FileName} not found for {folder}")
                            break
                        else:
                            cur_dir = parent_dir
            except FileNotFoundError as e:
                ExcelFile = None
                logging.error(f"Couldn't find folder {folder}. Error: {e}")
                not_found.append(folder)

            if ExcelFile:
                try:
                    MetadataDfRow = pd.read_excel(ExcelFile, header=0, index_col='Index')
                    ExcelFileDir = ExcelFile.rpartition('\\')[0].replace("\\", "/")
                    for column in ['Metadata File Location']:
                        if column not in gpdDf:
                            gpdDf[column] = ""
                            logging.info(f"Added metadata column {column}")
                    gpdDf.loc[gpdDf['SourceFileDir'] == folder, 'Metadata File Location'] = ExcelFileDir.replace("/", "\\")

                    for column in MetadataDfRow.index:
                        if column not in gpdDf:
                            gpdDf[column] = ""
                            logging.info(f"Added metadata column {column}")
                        gpdDf.loc[gpdDf['SourceFileDir'] == folder, column] = MetadataDfRow.loc[column, 'Values']
                except FileNotFoundError as e:
                    logging.error(f"Couldn't find metadata excel file in {folder}. Error:{e}")

        logging.info(f'Located and matched {fileNo} metadata files.')
        # a short script to remove folders that were missed - this might not be necessary in more recent version of the package
        filteredGdf = gpdDf
        if len(not_found):
            logging.info('Some folders are missing, will attempt to remove from database.')

            for folder in not_found:
                logging.debug(f"Removing following folder from database: {folder}")
                folder = folder.replace('\\', '/')
                filteredGdf = gpdDf[gpdDf['SourceFileDir'] != folder]
                removedRows = len(gpdDf) - len(filteredGdf)
                if removedRows:
                    logging.warning(f"Removed {removedRows} entries with folder value: {folder}")
                else:
                    logging.error(f'Error when removing following folder from database: {folder}')
        return filteredGdf

    def associate_metashape_projects(self, gpdDf):
        now = dt.now()
        if (os.path.exists(self.metashape_cache_file) and dt.fromtimestamp(os.path.getmtime(self.metashape_cache_file)) > now - td(minutes=self.metashapeCacheTime)):
            logging.info(f'Using cached Metashape project matches as time elapsed ({(now - dt.fromtimestamp(os.path.getmtime(self.metashape_cache_file))).total_seconds() / 60}) is less than cacheTime ({self.metashapeCacheTime}).')
            with open(self.metashape_cache_file, 'rb') as f:
                MSDataDict = pickle.load(f)
        else:
            MSDataDict = PsxLocator.main(self.MetashapeDirectory)
            with open(self.metashape_cache_file, 'wb') as f:
                pickle.dump(MSDataDict, f)
                logging.debug(f'Metashape data was saved to: {self.metashape_cache_file }')

        if MSDataDict is None or len(MSDataDict) == 0:
            logging.info("No Metashape project files contain any images from the database.")
            return gpdDf

        logging.info(f"{len(MSDataDict)} Metashape project file(s) contain images from the database.")

        MSDataDict = {k.lower().encode("utf-8"): v for k, v in MSDataDict.items()}
        gpdDf['SourceFileDirLower'] = gpdDf['SourceFileDir'].str.lower().str.encode("utf-8")

        for k in MSDataDict.keys():
            if k in gpdDf.SourceFileDirLower.values:
                logging.debug(f'Photos from folder {k} in current geodataframe')

        gpdDf['Metashape Files'] = gpdDf['SourceFileDirLower'].apply(lambda x: MSDataDict.get(x)).fillna('')
        gpdDf = gpdDf.drop(columns='SourceFileDirLower')

        return gpdDf



import pytest
import pandas as pd
from unittest.mock import patch, MagicMock
from exifScan.ExifToolScan import ExifToolScanner, process_output, reader
import os


def test_initialization():
    scanner = ExifToolScanner(FileTypes=['jpg', 'png'], recursive=True, ignoreDirs=['/ignore/this'], ignoreMetashape=True)
    assert scanner.FileTypes == ['jpg', 'png']
    assert scanner.recursive is True
    assert scanner.ignoreDirs == ['/ignore/this']
    assert scanner.ignoreMetashape is True


@patch('exifScan.ExifToolScan.subprocess.Popen')
@patch('exifScan.ExifToolScan.threading.Thread')
def test_run_exiftool(mock_thread, mock_popen):
    mock_process = MagicMock()
    mock_process.wait.return_value = None
    mock_popen.return_value = mock_process

    scanner = ExifToolScanner(FileTypes=['jpg'], recursive=False)
    result = scanner.run_exiftool('/path/to/scan')

    mock_popen.assert_called_once()
    assert result is not None


@patch('exifScan.ExifToolScan.os.path.getsize')
@patch('exifScan.ExifToolScan.pd.read_json')
def test_process_output(mock_read_json, mock_getsize):
    mock_getsize.return_value = 100
    mock_read_json.return_value = pd.DataFrame({'data': [1, 2, 3]})

    result = process_output('/path/to/scan', 'exifOut.json')
    assert not result.empty


@patch('exifScan.ExifToolScan.os.path.getsize')
@patch('exifScan.ExifToolScan.os.remove')
def test_process_output_empty_file(mock_remove, mock_getsize):
    mock_getsize.return_value = 0

    result = process_output('/path/to/scan', 'exifOut.json')
    assert result.empty
    mock_remove.assert_called_once_with('exifOut.json')


def test_reader():
    mock_pipe = MagicMock()
    mock_pipe.readline.side_effect = ['line1\n', 'line2\n', '']
    mock_func = MagicMock()
    stop_event = MagicMock()
    stop_event.is_set.return_value = False

    reader(mock_pipe, mock_func, stop_event)
    assert mock_func.call_count == 2


def test_images_in_folder():
    images_folder = 'tests/images'
    assert os.path.exists(images_folder)
    assert len(os.listdir(images_folder)) > 0  # Ensure there are images in the folder

    scanner = ExifToolScanner(FileTypes=['jpg', 'png'], recursive=True)
    result = scanner.run_exiftool(images_folder)

    # Ensure the result is a DataFrame
    assert isinstance(result, pd.DataFrame)

    # Check the number of rows matches the number of images
    num_images = len([f for f in os.listdir(images_folder) if f.lower().endswith(('jpg', 'png'))])
    assert len(result) == num_images

    # Verify specific columns
    assert 'SourceFile' in result.columns
    assert 'GPSLongitude' in result.columns
    assert 'GPSLatitude' in result.columns
    assert 'GPSAltitude' in result.columns
    assert 'CreateDate' in result.columns
    assert 'ImageSize' in result.columns
    assert 'FileSize' in result.columns
    assert 'Make' in result.columns
    assert 'Model' in result.columns
    # Validate data types
    assert result['SourceFile'].dtype == object
    assert result['GPSLongitude'].dtype == float
    assert result['GPSLatitude'].dtype == float
    assert result['GPSAltitude'].dtype == float
    assert result['CreateDate'].dtype == object
    assert result['ImageSize'].dtype == object
    assert result['FileSize'].dtype == 'int64'
    assert result['Make'].dtype == object
    assert result['Model'].dtype == object

    # Check for non-null values in important columns
    assert result['SourceFile'].notnull().all()
    assert result['GPSLongitude'].notnull().all()
    assert result['GPSLatitude'].notnull().all()
    assert result['GPSAltitude'].notnull().all()
    assert result['CreateDate'].notnull().all()
    assert result['ImageSize'].notnull().all()
    assert result['FileSize'].notnull().all()
    assert result['Make'].notnull().all()
    assert result['Model'].notnull().all()

    # Check that all 'Make' values are 'DJI'
    assert (result['Make'] == 'DJI').all()

    # Check that all 'Model' values are 'FC6310'
    assert (result['Model'] == 'FC6310').all()

    for date in result['CreateDate']:
        assert date.startswith('2024:07:09')
        assert len(date) == 19  # Ensure the format is 'YYYY:MM:DD HH:MM:SS'


def test_empty_directory():
    empty_folder = 'tests/test_folders/empty_images'
    os.makedirs(empty_folder, exist_ok=True)
    assert os.path.exists(empty_folder)
    assert len(os.listdir(empty_folder)) == 0  # Ensure the directory is empty

    scanner = ExifToolScanner(FileTypes=['jpg', 'png'], recursive=True)
    result = scanner.run_exiftool(empty_folder)
    assert result.empty  # Expect an empty DataFrame


def test_unsupported_file_types():
    unsupported_folder = 'tests/test_folders/unsupported_images'
    os.makedirs(unsupported_folder, exist_ok=True)
    with open(os.path.join(unsupported_folder, 'test.txt'), 'w') as f:
        f.write('This is a test file.')
    assert os.path.exists(unsupported_folder)
    assert len(os.listdir(unsupported_folder)) > 0  # Ensure there is at least one file

    scanner = ExifToolScanner(FileTypes=['jpg', 'png'], recursive=True)
    result = scanner.run_exiftool(unsupported_folder)
    assert result.empty  # Expect an empty DataFrame since no supported files are present


def test_corrupted_files():
    corrupted_folder = 'tests/test_folders/corrupted_images'
    os.makedirs(corrupted_folder, exist_ok=True)
    with open(os.path.join(corrupted_folder, 'corrupted.jpg'), 'wb') as f:
        f.write(b'\x00\x01\x02\x03\x04\x05')  # Write some invalid binary data
    assert os.path.exists(corrupted_folder)
    assert len(os.listdir(corrupted_folder)) > 0  # Ensure there is at least one file

    scanner = ExifToolScanner(FileTypes=['jpg', 'png'], recursive=True)
    result = scanner.run_exiftool(corrupted_folder)
    assert not result.empty  # Expect a non-empty DataFrame
    assert 'GPSLongitude' not in result.columns or result['GPSLongitude'].isnull().all()
    assert 'GPSLatitude' not in result.columns or result['GPSLatitude'].isnull().all()


def test_large_files():
    large_folder = 'tests/large_images'
    # Assuming you have a large image file for testing
    large_image_path = os.path.join(large_folder, 'CROWN_2853_B_3_cropped.tif')
    assert os.path.exists(large_image_path)

    scanner = ExifToolScanner(FileTypes=['tif', 'png'], recursive=True)
    result = scanner.run_exiftool(large_folder)
    assert not result.empty  # Expect a non-empty DataFrame


# def test_invalid_file_types():
#     invalid_folder = 'tests/test_folders/invalid_images'
#     os.makedirs(invalid_folder, exist_ok=True)
#     with open(os.path.join(invalid_folder, 'test.invalid'), 'w') as f:
#         f.write('This is a test file.')
#     assert os.path.exists(invalid_folder)
#     assert len(os.listdir(invalid_folder)) > 0  # Ensure there is at least one file

#     scanner = ExifToolScanner(FileTypes=['jpg', 'png'], recursive=True)
#     with pytest.raises(ValueError, match="Unsupported file type"):
#         scanner.run_exiftool(invalid_folder)


def test_non_existent_directory():
    non_existent_folder = 'tests/test_folders/non_existent_images'
    scanner = ExifToolScanner(FileTypes=['jpg', 'png'], recursive=True)
    result = scanner.run_exiftool(non_existent_folder)
    assert isinstance(result, pd.DataFrame)
    assert result.empty  # Expect a non-empty DataFrame


def test_invalid_json_format():
    invalid_json_folder = 'tests/test_folders/invalid_json'
    os.makedirs(invalid_json_folder, exist_ok=True)
    path = os.path.join(invalid_json_folder, 'invalid.json')
    with open(path, 'w') as f:
        f.write('{"invalid": "json"}')  # Write invalid JSON data
    assert os.path.exists(invalid_json_folder)
    assert len(os.listdir(invalid_json_folder)) > 0  # Ensure there is at least one file

    result = process_output(invalid_json_folder, path)
    assert result is None


def test_missing_required_parameters():
    with pytest.raises(TypeError, match="missing 1 required positional argument"):
        ExifToolScanner()  # Missing required parameters



from exifScan.metashape.PsxLocator import main
import sys
from unittest.mock import patch, MagicMock
import logging


def test_python_version():
    if sys.version_info[1] > 11:
        assert main("tests/images") is None

# test fails
# def test_metashape_import(caplog):
#     with patch.dict('sys.modules', {'Metashape': None}):
#         with caplog.at_level(logging.INFO):

#             assert main("tests/images") is None


def test_no_project_files_found(tmp_path):
    # Create a temporary directory
    d = tmp_path / "sub"
    d.mkdir()
    result = main(d)
    assert result is None


def test_project_files_found(tmp_path, caplog):
    # Create a temporary directory with project files
    d = tmp_path / "sub"
    d.mkdir()
    (d / "project1.psx").write_text("content")
    (d / "project2.psz").write_text("content")

    with patch('Metashape.Document.open', return_value=None):
        with patch('Metashape.Document') as MockDocument:

            mock_doc = MockDocument.return_value
            mock_chunk = MagicMock()
            mock_camera = MagicMock()
            mock_camera.photo.path = "/some/path/photo.jpg"
            mock_chunk.cameras = [mock_camera]
            mock_doc.chunks = [mock_chunk]
            mock_doc.chunks = [mock_chunk]

            result = main(d)
            assert result is not None
            assert len(result) == 1
            assert "/some/path" in result


def test_camera_without_photo(tmp_path):
    # Create a temporary directory with project files
    d = tmp_path / "sub"
    d.mkdir()
    (d / "project1.psx").write_text('<?xml version="1.0" encoding="UTF-8"?><document version="1.2.0" path="{projectname}.files/project.zip"/>')

    with patch('Metashape.Document') as MockDocument:
        mock_doc = MockDocument.return_value
        mock_chunk = MagicMock()
        mock_camera = MagicMock()
        mock_camera.photo = None
        mock_chunk.cameras = [mock_camera]
        mock_doc.chunks = [mock_chunk]

        result = main(d)
        assert result is not None
        assert len(result) == 0


def test_with_images(tmp_path, caplog):
    # this whole path business is a mess..

    # Create a temporary directory with project files and images
    d = tmp_path / "sub"
    d.mkdir()
    (d / "project1.psx").write_text("content")
    (d / "project2.psz").write_text("content")

    # Create images directory
    images_dir = d / "tests/images"
    images_dir.mkdir(parents=True)
    (images_dir / "image1.jpg").write_text("image content")
    (images_dir / "image2.jpg").write_text("image content")

    with patch('Metashape.Document.open', return_value=None):
        with patch('Metashape.Document') as MockDocument:
            mock_doc = MockDocument.return_value
            mock_chunk = MagicMock()
            mock_camera1 = MagicMock()
            mock_camera1.photo.path = str(images_dir / "image1.jpg").replace("\\", "/")
            mock_camera2 = MagicMock()
            mock_camera2.photo.path = str(images_dir / "image2.jpg").replace("\\", "/")
            mock_chunk.cameras = [mock_camera1, mock_camera2]
            mock_doc.chunks = [mock_chunk]

            result = main(d)
            logging.info(result)
            sanitised_images_dir = str(images_dir).replace("\\", "/")
            assert result is not None
            assert len(result) == 1
            assert sanitised_images_dir in result
            assert "project1.psx" in result[str(images_dir).replace("\\", "/")]


def test_with_test_project(tmp_path, caplog):

    test_project_path = 'tests/images/test.psx'
    test_path = 'tests/images'
    with caplog.at_level(logging.INFO):
        result = main(test_path)

        logging.info(result)
        assert result is not None
        assert len(result) == 1
        assert test_path in list(result.keys())[0]
        assert test_project_path in list(result.values())[0]



# a separate module with functions that generate snippets to make the main HTML function more legible.
import json
import hashlib


def stable_hash(name):
    return int(hashlib.md5(name.encode()).hexdigest(), 16)


def cssColors_for_geojson(name, color_list, mapElementsDict):
    # Create a color from a stable hash of name
    index = stable_hash(name) % len(color_list)
    color = color_list.pop(index)
    mapElementsDict[name]['color'] = color
    el_id = name.replace(' ', '_')

    return f'''
.{el_id}{{
    stroke:{color};
    fill:{color};
}}

'''


def HTMLTail(navlinks, tabsContent, jsContent, copyButtonString):
    html = '''
<div class="offcanvas offcanvas-bottom" data-bs-scroll="true" data-bs-backdrop="false" tabindex="-1" id="offcanvasScrolling" aria-labelledby="offcanvasScrollingLabel">
  <div class="offcanvas-header">
    <h5 class="offcanvas-title" id="offcanvasScrollingLabel"></h5>
    <button type="button" class="btn-close" data-bs-dismiss="offcanvas" aria-label="Close"></button>
  </div>
  <div id="offcanvas-body-id" class="offcanvas-body"><div class="nav nav-tabs" id="nav-tab" role="tablist">'''
    html = html + ''.join(navlinks)
    html = html + '</div><div class = "tab-content" id = "tableTabContent" >'
    html = html + ''.join(tabsContent)
    html = html + '</div>  </div></div><script>$(document).ready( function () {'
    html = html + ''.join(jsContent)
    html = html + f' }} );</script><script>{copyButtonString}</script></html>'
    return html


def rightClick(mapElementsDict):
    colors = {}
    for key, value in mapElementsDict.items():
        if 'color' in value:
            colors[key] = value['color']
    colors_json = json.dumps(colors)

    return f'''
<script>
    const colors = {colors_json}
    class ElementToggler {{
        constructor(targetClass, toggleClass) {{
            this.targetClass = targetClass;
            this.toggleClass = toggleClass;
            this.contextEvents = new Map();

        }}

        addEvent() {{
            let elements = document.getElementsByClassName(this.targetClass);
            for (let i = 0; i < elements.length; i++) {{
                const element = elements[i];

                if (!this.contextEvents.has(element)) {{
                    element.addEventListener('contextmenu', (event) => {{
                        event.preventDefault();
                        element.classList.toggle(this.toggleClass);

                    }});
                    this.contextEvents.set(element, true);

                }}

            }}

        }}

    }}

    class ChangeLabelColors{{
         constructor(targetClass) {{
            this.targetClass = targetClass;
        }}
        changeColors(){{
            var checkboxes = document.querySelectorAll(this.targetClass);
            for (var i = 0; i < checkboxes.length; i++) {{
                const checkbox= checkboxes[i];
                const name =checkbox.nextSibling.innerText.trim();
                const color = colors[name];
                checkbox.parentElement.style.color = color;

            }}

        }}


    }}

    const toggler = new ElementToggler('animateVisibility', 'noFill');
    const colorsetup = new ChangeLabelColors('.leaflet-control-layers input[type="checkbox"]');

    window.onload = function() {{
        //add context menu event on click
        const control = document.getElementsByClassName('folium-map')[0]
        control.addEventListener('pointerup',()=>{{
            toggler.addEvent();
            // set timeout to ensure popup is populated.
            setTimeout(makeCopyable,400);


        }});

        colorsetup.changeColors();


    }};

    const copyMap = new WeakMap();

    function makeCopyable(){{
        const copyEL = document.getElementsByClassName('copyText');
        for (var i = 0; i < copyEL.length; i++) {{
            const parent = copyEL[i].parentElement;
            const cellText= parent.textContent;
            if(!copyMap.has(parent)){{

                parent.addEventListener( 'click', () => {{

                    // Copy the cell text to clipboard
                    navigator.clipboard.writeText( cellText )
                        .then( () => {{

                            console.log( `Copied: ${{cellText
                            }}` );
                            parent.classList.add('flashGreen');

                            // Remove the 'flash' class after the animation has completed
                            setTimeout(function() {{
                                parent.classList.remove('flashGreen');

                            }}, 600); // The same duration as your animation


                        }} )
                        .catch( error => {{

                            console.error( 'Error copying text:', error );
                             parent.classList.add('flashRed');

                            // Remove the 'flash' class after the animation has completed
                            setTimeout(function() {{
                                parent.classList.remove('flashRed');

                            }}, 200); // The same duration as your animation

                        }} );

                }} );
                copyMap.set(parent,true);

            }}

        }}

    }}
</script>
'''


def copyButton():

    return '''

function makeCellsCopyable(tableId, headerName) {
    const table = document.getElementById(tableId);
    if (table) {
        const headerCells = table.querySelectorAll("thead tr th");

        headerCells.forEach((headerCell, index) => {
            const columnHeader = headerCell.innerHTML; // Get the header name

            if (columnHeader === headerName) {
                const dataCells = table.querySelectorAll(
                    `tbody tr td:nth-child(${index + 1})`,
                );
                dataCells.forEach((dataCell) => {
                    const cellText = dataCell.textContent;
                    dataCell.textContent = "";

                    // Create a div element
                    const div = document.createElement("div");
                    div.style.display = "flex";
                    div.style.justifyContent = "space-between";
                    div.style.alignItems = "flex-start";

                    // Create a text node for the cell text
                    const textNode = document.createTextNode(cellText);
                    div.appendChild(textNode);

                    // Create a button element
                    const copyButton = document.createElement("button");
                    copyButton.classList.add("btn", "btn-secondary", "d-flex");
                    copyButton.type = "button";
                    copyButton.innerHTML = `<svg xmlns="http://www.w3.org/2000/svg" width="12" height="12" fill="currentColor" class="bi bi-clipboard" viewBox="0 0 16 16">
                        <path d="M4 1.5H3a2 2 0 0 0-2 2V14a2 2 0 0 0 2 2h10a2 2 0 0 0 2-2V3.5a2 2 0 0 0-2-2h-1v1h1a1 1 0 0 1 1 1V14a1 1 0 0 1-1 1H3a1 1 0 0 1-1-1V3.5a1 1 0 0 1 1-1h1z"/>
                        <path d="M9.5 1a.5.5 0 0 1 .5.5v1a.5.5 0 0 1-.5.5h-3a.5.5 0 0 1-.5-.5v-1a.5.5 0 0 1 .5-.5zm-3-1A1.5 1.5 0 0 0 5 1.5v1A1.5 1.5 0 0 0 6.5 4h3A1.5 1.5 0 0 0 11 2.5v-1A1.5 1.5 0 0 0 9.5 0z"/>
                    </svg>`;
                    div.appendChild(copyButton);

                    // Append the div to the cell
                    dataCell.appendChild(div);

                    copyButton.addEventListener("click", (e) => {
                        e.stopPropagation();

                        // Copy the cell text to clipboard
                        navigator.clipboard
                            .writeText(cellText)
                            .then(() => {
                                console.log(`Copied: ${cellText}`);
                            })
                            .catch((error) => {
                                console.error("Error copying text:", error);
                            });
                    });
                    // Append the button to the cell
                });
            }
        });
    }
}
'''


def navLink(navId, name, selected=False):
    active = '' if selected is False else ' active'
    return f'<button class="nav-link{active}" id="{navId}-tab" data-bs-toggle="tab" data-bs-target="#target{navId}" type="button" role="tab" aria-controls="{navId}" aria-selected="true">{name}</button>'


def tabDivs(navId, name, content, selected=False):
    classString = '' if selected is False else 'show active'

    return f'''
    <div class = "tab-pane fade {classString}" id = "target{navId}" role = "tabpanel" aria-labelledby="{navId}-tab">
    <h2 class='text-center p-3'>{name}</h2>
    {content.to_html(table_id="table" + navId, classes="table table-striped responsive")}
    </div>
    '''


def tableJs(navId):
    return f"$('#table{navId}').DataTable({{responsive:true}});  document.getElementById('table{navId}').style.width='';"


def zoomOnEl(soup, mapId, tableId, geojsonId):
    script = soup.new_tag('script', type='module')
    script.append(f'''const table = document.getElementById( '{tableId}' );
    const headerCells = table.querySelectorAll( 'thead tr th' );
    headerCells.forEach( ( headerCell, index ) => {{
        const columnHeader = headerCell.innerHTML; // Get the header name
        if ( columnHeader == 'index' ) {{
            const dataCells = table.querySelectorAll( `tbody tr td:nth-child(${{index + 1}})` );
            dataCells.forEach(
                dataCell => {{
                    dataCell.parentElement.addEventListener('click',()=>zoomToFeature({geojsonId},dataCell.textContent));
                }}
             );
        }}
    }} );
function zoomToFeature(geojsonId,index) {{
    if( geojsonId && {mapId}){{
        geojsonId.eachLayer(function (layer) {{
            if (layer.feature.properties.index == index) {{
                {mapId}.addLayer(geojsonId)
                {mapId}.flyToBounds(layer.getBounds(),{{duration:2}});
            }}
        }});
    }}
}}
''')
    return script



# This script is made to be run from
# \\file\Shared\SEESPostGrad\JD\Geospatial database\scripts
# and has ExifScanJSON.py, PsxLocator.py and CreateMetadata.py
# as dependencies. As well as the obvious python dependencies,
# it also needs the Metashape script api installed with a
# functioning license in the folder, license in the host@port format.
# JD 23/01/2020

from datetime import datetime as dt
import os

import geopandas as gpd
import pandas as pd
import subprocess
import pyogrio
import logging
from exifScan.CreateMetadata import MetadataWriter
from exifScan.ExifToolScan import ExifToolScanner
from exifScan.geoDfFromExifDf import GeoDataProcessor
from exifScan.generateHTML.createHTML import HTMLWriter

# https://geopandas.org/en/stable/docs/user_guide/io.html
gpd.options.io_engine = "pyogrio"


class ExifScanner:
    # should refactor this with *args,**kwargs patterns now that we have defaults in the __main__ part. Same for all helper classes
    def __init__(self, WorkDirectory=None, MetashapeDirectory=None, KartDirectory=None, KartRemote=None,
                 GeopackageDirectory=None, OutputGeopackage='Photos.gpkg', htmlFolder=None, mapFileName='map.html',
                 otherMapLayers=None, otherMapLayerNames=None, ExifOutputFile="exiftoolOut",
                 readExifFromOutputFile=False, recursive=False, ignoreDirs=[], ignoreMetashape=False,
                 groupByTags=["Make", "Model"], xyTagNames=['GPSLongitude', 'GPSLatitude'], zTagName='GPSAltitude',
                 imageFileExtensions=['JPG'], metashapeCacheTime=120, stats={'Scan Started at': "", }, linzAPIKey=None
                 ):

        self.WorkDirectory = WorkDirectory or os.getcwd()
        self.MetashapeDirectory = MetashapeDirectory
        self.KartDirectory = KartDirectory
        self.KartRemote = KartRemote
        self.GeopackageDirectory = GeopackageDirectory or self.WorkDirectory
        self.OutputGeopackage = OutputGeopackage
        self.htmlFolder = htmlFolder
        self.mapFileName = mapFileName
        self.otherMapLayers = otherMapLayers
        self.otherMapLayerNames = otherMapLayerNames
        self.ExifOutputFile = ExifOutputFile
        self.readExifFromOutputFile = readExifFromOutputFile
        self.recursive = recursive
        self.ignoreDirs = ignoreDirs
        self.ignoreMetashape = ignoreMetashape
        self.groupByTags = groupByTags
        self.xyTagNames = xyTagNames
        self.zTagName = zTagName
        self.imageFileExtensions = [s.upper() for s in imageFileExtensions]
        self.metashapeCacheTime = metashapeCacheTime
        self.splitSize = 100
        self.OutputGeopackagePath = os.path.join(self.GeopackageDirectory, self.OutputGeopackage)
        self.non_geoloacted_path = os.path.join(self.WorkDirectory, 'nongeolocated.csv')
        self.gpkg_layer = 'allGroupedData'
        self.metadata_writer = MetadataWriter(self.WorkDirectory, self.metashapeCacheTime, self.MetashapeDirectory)
        self.exifToolScanner = ExifToolScanner(self.imageFileExtensions, self.recursive, self.ignoreDirs, self.ignoreMetashape)
        self.geoprocessor = GeoDataProcessor(WorkDirectory, xyTagNames, zTagName, self.groupByTags)
        self.linzAPIKey = linzAPIKey
        self.htmlWriter = HTMLWriter(self.OutputGeopackagePath, self.htmlFolder, self.mapFileName, self.otherMapLayers, self.otherMapLayerNames, self.splitSize, self.linzAPIKey)
        self.stats = stats

    def write_HTML(self):
        if self.htmlFolder:
            # set leaflet map and table location
            info = ''
            for feature, value in self.stats.items():
                info = info + f'<span>{feature}: {value}</span><br/>'

            self.htmlWriter.updateHTML(info)

    def scan_for_photos(self, ScanFolder):
        StartTime = dt.now()
        now = StartTime.strftime("%Y%m%d%H%M%S")
        startCtime = StartTime.ctime()

        if not ScanFolder:
            raise Exception("Error: ScanFolder is not defined or is False")
        logging.info(f'Started folder Scan at {startCtime}')

        logging.debug(f'Current working directory is: {self.WorkDirectory}')

        if not self.MetashapeDirectory:
            self.metadata_writer.MetashapeDirectory = ScanFolder
            logging.debug(f'Current Metashape directory is: {ScanFolder}')

        ExifStartTime = dt.now()
        now = ExifStartTime.strftime("%Y%m%d%H%M%S")

        if self.readExifFromOutputFile:
            ExifOutputFileName = self.ExifOutputFile + '.json'
            exifFile = os.path.join(self.WorkDirectory, ExifOutputFileName)
            df = pd.read_json(exifFile)
        else:
            ExifOutputFileName = self.ExifOutputFile + str(now) + '.json'
            exifFile = os.path.join(self.WorkDirectory, ExifOutputFileName)
            df = self.exifToolScanner.run_exiftool(ScanFolder, exifFile)
        if df is None:
            logging.error('Error when running ExifScan')
            return None

        logging.info("Exiftool done processing. %s photos were found and exif data was extracted. The tool process took %s (h:mm:ss).", len(df), dt.now() - ExifStartTime)

        if df.empty:
            if not self.recursive:
                logging.info('Scanned folder has no images, will attempt to remove from database.')
                result = self.removefromGdf(ScanFolder)
                if not result:
                    logging.error('Error when removing from database, aborting script')
                    return
                self.update_data()
            logging.info('Script ended as no images found')
            return ScanFolder

        logging.info('Processing data using pandas and geopandas.')

        geoDfStartTime = dt.now()
        points, geoDf, nonGeolocated = self.geoprocessor.process(df)

        if len(nonGeolocated):
            logging.warning(f'Found {len(nonGeolocated)} nonGeolocated group(s) in the folder with exention(s) {self.imageFileExtensions}.')

            nonGeolocated = self.metadata_writer.update_metadata(nonGeolocated)

            try:
                with open(self.non_geoloacted_path, 'a') as f:
                    f.write(f'ScanFolder: {ScanFolder} time: {dt.now() - geoDfStartTime}\n')
                nonGeolocated.to_csv(self.non_geoloacted_path, mode='a', index=False)
            except Exception as e:
                logging.exception(f"An unexpected error occurred: {e}. The error type is : {e.__class__.__name__}")

        logging.info("Geopandas done processing. The tool process took %s (h:mm:ss).", dt.now() - geoDfStartTime)

        if points is None or len(points) == 0:
            if not self.recursive:
                logging.info('Scanned folder has no geo-located images, will attempt to remove folder from database.')
                result = self.removefromGdf(ScanFolder)
                if not result:
                    logging.error('Error when removing from database, aborting script')
                    return
                self.update_data()
            logging.info('Script ended as no geolocated images identified')
            return ScanFolder

        logging.info('Creating geopackages and tables.')

        tost = geoDf.copy()
        tost = tost.to_crs(3857)
        invalidGeometries = tost.geometry.apply(lambda g: not g.is_valid)
        if len(tost[invalidGeometries]):
            logging.warning(f'There are {len(tost[invalidGeometries])} invalid geometries in the geodataframe when projected in "Web Mercator".')
            logging.debug(tost[invalidGeometries])

        geoDf['areaSqkm'] = tost.area / 10**6

        gpkgname = 'Database' + now + '.gpkg'
        gpkgPath = os.path.join(self.WorkDirectory, gpkgname)
        points.to_file(gpkgPath, layer='points', driver="GPKG")

        if len(geoDf):
            geoDf.to_file(gpkgPath, layer=self.gpkg_layer, driver="GPKG")
            logging.debug(f"Archived filtered polygon called: allGroupedData in {gpkgPath}. It has {len(geoDf)} feature(s).")
            self.replaceInGPKG(geoDf, ScanFolder)

        self.update_data()

        logging.info(f"Created files from the photo data. The whole thing took {dt.now() - StartTime} (h:mm:ss).")
        return ScanFolder

    def update_data(self):
        if self.KartDirectory:
            kartCmd = [
                "kart",
                "import",
                self.OutputGeopackagePath,
                "--all-tables",
                "--replace-existing"
            ]
            try:
                result = subprocess.run(kartCmd, cwd=self.KartDirectory)
                if result.returncode == 41:
                    logging.warning("Kart repository has not been initialised. Initialising now...")
                    init_process = subprocess.run(['kart', 'init', '--import', self.OutputGeopackagePath], cwd=self.KartDirectory)

                    if init_process.returncode == 0:
                        logging.info("Kart repository initialised successfully.")
                    else:
                        logging.error(f"Failed to initialise Kart repository. Return code: {init_process.returncode}")
                elif result.returncode == 44:
                    logging.info("Kart repository recorded no changes.")
                elif result.returncode == 0:
                    logging.info("Kart repository imported successfully.")

                    if self.KartRemote:
                        remoteResult = subprocess.run(['kart', 'push'], cwd=self.KartDirectory)

                        if remoteResult.returncode == 128:
                            logging.warning("Kart remote repository has not been set. Setting now...")
                            remoteAddResult = init_process = subprocess.run(['kart', 'remote', 'add', 'origin', self.KartRemote], cwd=self.KartDirectory)
                            pushResult = init_process = subprocess.run(['kart', 'push', '--set-upstream', 'origin', 'main'], cwd=self.KartDirectory)
                            if remoteResult.returncode == 0 and pushResult == 0:
                                logging.info(f"Kart {self.KartRemote} remote repository set successfully.")
                            else:
                                logging.error(f"Failed to set Kart remote repository.\nRemote add return code: {remoteAddResult.returncode}\n Push result:{pushResult.returncode}")
                        elif remoteResult.returncode == 0:
                            logging.info('Kart repository pushed successfully.')
                        else:
                            logging.error(f'Kart remote failed with error code : {remoteResult.returncode}')
                else:
                    logging.error(f'Kart import failed with error code : {result.returncode}')

            except NotADirectoryError:
                logging.error(f'kart directory doesn\'t exist:{self.KartDirectory} .Data will not be backed up to a repository.')

        # Create a map centered at the mean coordinates of your geodata

        self.write_HTML()

    def replaceInGPKG(self, geoDf, ScanFolder):

        try:
            # Read the existing data from a GeoPackage file
            existing = gpd.read_file(self.OutputGeopackagePath, layer=self.gpkg_layer, fid_as_index=True)

            # Determine the maximum index in the existing dataframe
            max_index = existing.index.max()

            # Set the index of the new GeoDataFrame to start after the largest index in the existing dataframe
            geoDf.index = geoDf.index + max_index + 1

            # Add metadata to the new GeoDataFrame for duplicate checking
            geoDf = self.metadata_writer.update_metadata(geoDf)

            # Concatenate the existing and new GeoDataFrames
            concatDf = pd.concat([existing, geoDf])

            # If not in recursive mode, remove rows that are not in the new GeoDataFrame (indicating modifications)
            if not self.recursive:
                # Filter existing rows to keep only those from the specified folder
                existingFiltered = existing[existing['SourceFileDir'] == ScanFolder.replace('\\', '/')]
                # Keep only relevant columns for comparison
                existingFiltered = existingFiltered.copy()[self.groupByTags + ['SourceFileDir']]
                newValuesFiltered = geoDf.copy()[self.groupByTags + ['SourceFileDir']]
                # Identify rows unique to the existing dataset
                uniqueToExisting = existingFiltered[~existingFiltered.apply(tuple, 1).isin(newValuesFiltered.apply(tuple, 1))]
                # Remove those rows from the concatenated dataframe
                concatDf = concatDf.drop(uniqueToExisting.index)
                if len(uniqueToExisting):
                    logging.info(f'Removed {len(uniqueToExisting)} entries that were no longer present in scanned folder.')

            # Keep the first occurrence of entirely identical rows
            filteredConcatDf = concatDf.copy().drop_duplicates(keep='first')

            # Sort the DataFrame by specified columns
            filteredConcatDf.sort_values(by=[*self.groupByTags, 'SourceFileDir'], inplace=True)

            # Create a dictionary where keys are indices of the first occurrence of each duplicate set
            firstOccurrenceIndices = filteredConcatDf.drop_duplicates([*self.groupByTags, 'SourceFileDir'], keep='first')

            # Get the indices of the first occurrence of each duplicate set
            firstIndices = firstOccurrenceIndices.index

            # Drop duplicates and keep the last occurrence
            filteredConcatDf = filteredConcatDf.drop_duplicates([*self.groupByTags, 'SourceFileDir'], keep='last')

            # Create a new DataFrame with indices of the first occurrence of each duplicate set
            newDf = filteredConcatDf.copy()
            newDf.index = firstIndices
            newDf.index.name = 'fid'

        except pyogrio.errors.DataSourceError as e:
            logging.warning(f"Error: {e}. This is normal is this is the first recursive Scan. A new df will be created.")
            newDf = geoDf
            newDf.index.name = 'fid'
        except pyogrio.errors.DataLayerError:
            logging.warning(f"Layer {self.gpkg_layer} doesn't exist. This is normal is this is the first time creating this layer. A new df will be created.")
            newDf = geoDf
            newDf.index.name = 'fid'

        # update metaData on all rows in df
        # doing this additional update as it is relatively quick if the Metashape values are cached.
        newDf = self.metadata_writer.update_metadata(newDf)

        newDf.to_file(self.OutputGeopackagePath, layer=self.gpkg_layer, driver="GPKG")
        logging.info(f"Replaced or wrote {self.gpkg_layer} dataset to the '{self.gpkg_layer}' layer in {self.OutputGeopackagePath},which now has a total of {len(newDf)} feature(s).")

    def removefromGdf(self, value):
        try:  # check if file exists before continuing
            gpd.read_file(self.OutputGeopackagePath)
        except pyogrio.errors.DataSourceError as e:
            logging.error(f"Error: {e}. Check if the file exists or if the path is correct.")
            return True
        except Exception as e:
            logging.exception(f"An unexpected error occurred: {e}. The error type is : {e.__class__.__name__}")

        layers = pyogrio.list_layers(self.OutputGeopackagePath)

        for layer in layers:
            # layer is in format ['name' 'type']
            layerName = layer[0]
            existing = gpd.read_file(self.OutputGeopackagePath, layer=layerName, fid_as_index=True)
            # keep everything in forward slashes.
            value = value.replace('\\', '/')

            filteredGdf = existing[existing['SourceFileDir'] != value]
            # Debugging statements
            filteredGdf.to_file(self.OutputGeopackagePath, layer=layerName, driver="GPKG")
            removedRows = len(existing) - len(filteredGdf)
            if removedRows:
                logging.warning(f"Removed {removedRows} rows with value :{value} in {self.OutputGeopackagePath}")
        return True
